#define _WIN32_WINNT 0x0400

#include <windows.h>
#include <stdio.h>
#include "stdserv.h"

#define EXT2KEYUSERS "Software\\Utilities\\Ext2Tools\\Users"
#define EXT2KEYPARAMS "Software\\Utilities\\Ext2Tools\\Parameters"
#define PARAM_AUTOTERMINATE "AutoTerminate"

#ifdef SHOWMESSAGES
#define MESSAGEBOX(MSG) MessageBox(NULL,MSG,CP_SERVICE_NAME,MB_SERVICE_NOTIFICATION)
#else
#define MESSAGEBOX(MSG)
#endif

BOOL ServiceInitialized=FALSE;
HANDLE ActionEvents[2];
PHANDLE ChangeNotificationEvent=ActionEvents;
PHANDLE StopEvent=ActionEvents+1;

BOOL GetAccountSid(LPTSTR SystemName, 
                   LPTSTR AccountName, 
                   PSID *Sid) 
{ 
    LPTSTR ReferencedDomain=NULL; 
    DWORD cbSid=128;    // initial allocation attempt 
    DWORD cchReferencedDomain=16; // initial allocation size 
    SID_NAME_USE peUse; 
    BOOL bSuccess=FALSE; // assume this function will fail 
 
    // 
    // initial memory allocations 
    // 
    *Sid = (PSID)HeapAlloc(GetProcessHeap(), 0, cbSid); 
    if (*Sid == NULL) goto cleanup;
 
    ReferencedDomain = (LPTSTR)HeapAlloc(GetProcessHeap(), 
                                         0, 
                                         cchReferencedDomain * sizeof(TCHAR)); 
    if (ReferencedDomain == NULL) goto cleanup; 
 
    // 
    // Obtain the SID of the specified account on the specified system. 
    // 
    while (!LookupAccountName(SystemName,         // machine to lookup account on 
                             AccountName,        // account to lookup 
                             *Sid,               // SID of interest 
                             &cbSid,             // size of SID 
                             ReferencedDomain,   // domain account was found on 
                             &cchReferencedDomain, 
                             &peUse)) { 
        if (GetLastError() == ERROR_INSUFFICIENT_BUFFER) { 
            // 
            // reallocate memory 
            // 
            *Sid = (PSID)HeapReAlloc(GetProcessHeap(), 
                                     0, 
                                     *Sid, 
                                     cbSid); 
            if (*Sid == NULL) goto cleanup; 
 
            ReferencedDomain = (LPTSTR)HeapReAlloc(GetProcessHeap(), 
                                                   0, 
                                                   ReferencedDomain, 
                                                   cchReferencedDomain * sizeof(TCHAR)); 
            if(ReferencedDomain == NULL) goto cleanup; 
        } 
        else goto cleanup; 
    } 
 
    // 
    // Indicate success. 
    // 
    bSuccess = TRUE; 
 
cleanup:
 
    // 
    // Cleanup and indicate failure, if appropriate. 
    // 
 
    HeapFree(GetProcessHeap(), 0, ReferencedDomain); 
 
    if(!bSuccess) { 
        if(*Sid != NULL) { 
            HeapFree(GetProcessHeap(), 0, *Sid); 
            *Sid = NULL; 
        } 
    }
 
    return bSuccess; 
} 

#define ADDACE(pAcl,AccessMask,pSid)             \
    if (!AddAccessAllowedAce(pAcl,               \
                             ACL_REVISION,       \
                             AccessMask,         \
                             pSid)) {            \
        printf("AddAccessAllowedAce failed.\n"); \
    }

BOOL GetNextValidSid(UCHAR **p,PSID *pSid)
{
    UCHAR *q;

    // skip double ','
    while (**p==',') *p += 1;

    q=strchr(*p,','); // search next ',' (last character in *p is a ',')
    if (q!=NULL) {
        *q='\0'; // replace ',' with termination character
        printf("\t\tExamining entry %s...",*p);
        if (!GetAccountSid(NULL,*p,pSid)) {
            printf("invalid, skipping.\n");
            *p=q+1; // move to character after ','
            return GetNextValidSid(p,pSid);
        }
        printf("ok.\n");
        *p=q+1; // move to character after ','
        return TRUE;
    } else {
        return FALSE;
    }
}

BOOL ApplyDACLs(void)
{
    HKEY                     hKey;
    DWORD                    dwIndex=0;
    UCHAR                    ValueName[MAX_PATH];
    DWORD                    cbValueName=MAX_PATH;
    DWORD                    Type;
    UCHAR                    Data[MAX_PATH];
    DWORD                    cbData=MAX_PATH;
    UCHAR                    NameList[MAX_PATH];
    UCHAR                    *p;
    UCHAR                    DeviceName[MAX_PATH];
    PSID                     pSidUser;
    PSID                     pSidAdministrators;
    PSID                     pSidSystem;
    PSID                     pSidEveryone;
    UCHAR                    SDBuf[SECURITY_DESCRIPTOR_MIN_LENGTH];
    PSECURITY_DESCRIPTOR     pSD=(PSECURITY_DESCRIPTOR)SDBuf;
    PACL                     pDACL;
    DWORD                    dwDACLBufSize;
    SID_IDENTIFIER_AUTHORITY siaNtAuthority = SECURITY_NT_AUTHORITY;
    SID_IDENTIFIER_AUTHORITY siaWorldSidAuthority = SECURITY_WORLD_SID_AUTHORITY;
    BOOL                     IsRemovable;
    BOOL                     bSuccess=FALSE;

    // create SIDs for Administrators, System and Everyone
    AllocateAndInitializeSid(&siaNtAuthority,
                             2,
                             SECURITY_BUILTIN_DOMAIN_RID,
                             DOMAIN_ALIAS_RID_ADMINS,
                             0,0,0,0,0,0, // six unused subauthorities
                             &pSidAdministrators);
    AllocateAndInitializeSid(&siaNtAuthority,
                             1,
                             SECURITY_LOCAL_SYSTEM_RID,
                             0,0,0,0,0,0,0, // seven unused subauthorities
                             &pSidSystem);
    AllocateAndInitializeSid(&siaWorldSidAuthority,
                             1,
                             SECURITY_WORLD_RID,
                             0,0,0,0,0,0,0, // seven unused subauthorities
                             &pSidEveryone);

    if (RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                     TEXT(EXT2KEYUSERS),
                     0,
                     KEY_ALL_ACCESS,
                     &hKey)!=ERROR_SUCCESS) {
        printf("Error opening registry key.\n");
        goto cleanup;
    }

    printf("Scanning registry key HKEY_LOCAL_MACHINE\\%s...\n\n",EXT2KEYUSERS);
    
    while (RegEnumValue(hKey,         // handle of key to query 
                        dwIndex++,    // index of value to query 
                        ValueName,    // address of buffer for value string 
                        &cbValueName, // address for size of value buffer 
                        NULL,         // reserved 
                        &Type,        // address of buffer for type code 
                        Data,         // address of buffer for value data 
                        &cbData       // address for size of data buffer 
                       )==0) { // !=ERROR_NO_MORE_ITEMS) {
        printf("Drive %s has user/group list %s.\n",ValueName,Data);

        IsRemovable=GetDriveType(ValueName)==DRIVE_REMOVABLE;
        if (IsRemovable) {
            dwDACLBufSize=sizeof(ACCESS_ALLOWED_ACE)-sizeof(DWORD)
                          +GetLengthSid(pSidEveryone)+sizeof(ACL);
        } else {
            dwDACLBufSize=3*(sizeof(ACCESS_ALLOWED_ACE)-sizeof(DWORD))
                          +GetLengthSid(pSidAdministrators)+GetLengthSid(pSidSystem)
                          +GetLengthSid(pSidEveryone)+sizeof(ACL);
        }

        printf("\tDetermining size for DACL buffer.\n");
        strcpy(NameList,Data);
        p=NameList;
        strcat(NameList,",");
        while (GetNextValidSid(&p,&pSidUser)) {
            dwDACLBufSize += sizeof(ACCESS_ALLOWED_ACE)
                            +GetLengthSid(pSidUser)-sizeof(DWORD);
            HeapFree(GetProcessHeap(),0,pSidUser);
        }
        printf("\tDACL size calculated as %i bytes.\n",dwDACLBufSize);

        printf("\tCreating DACL and adding ACEs.\n");
        pDACL=HeapAlloc(GetProcessHeap(),0,dwDACLBufSize);
        if (!InitializeAcl(pDACL,
                           dwDACLBufSize,
                           ACL_REVISION)) {
            printf("InitializeAcl failed.\n");
            return FALSE;
        }

        if (IsRemovable) {
            printf("\t\tAdding default ACE: Everyone/Full.\n");
            ADDACE(pDACL,GENERIC_ALL,pSidEveryone);
        } else {
            printf("\t\tAdding default ACE: Administrators+System/Full, Everyone/X.\n");
            ADDACE(pDACL,GENERIC_ALL,pSidAdministrators);
            ADDACE(pDACL,GENERIC_ALL,pSidSystem);
            ADDACE(pDACL,GENERIC_EXECUTE,pSidEveryone);
        }

        strcpy(NameList,Data);
        p=NameList;
        strcat(NameList,",");
        while (GetNextValidSid(&p,&pSidUser)) {
            ADDACE(pDACL,GENERIC_READ|GENERIC_EXECUTE,pSidUser);
            HeapFree(GetProcessHeap(),0,pSidUser);
        }

        InitializeSecurityDescriptor(pSD,SECURITY_DESCRIPTOR_REVISION);

        SetSecurityDescriptorDacl(pSD,
                                  TRUE,      // fDaclPresent flag
                                  pDACL,
                                  FALSE);    // not a default DACL

        if (IsValidSecurityDescriptor(pSD)) {
            sprintf(DeviceName,"\\\\.\\%s",ValueName);
            if (SetFileSecurity(DeviceName,
                                (SECURITY_INFORMATION)(DACL_SECURITY_INFORMATION),
                                pSD)==0) {
                printf("Error setting DACL for %s.\n",DeviceName);
            }
        } else {
            printf("Error, invalid SD\n");
        }
        HeapFree(GetProcessHeap(),0,pDACL);
        cbValueName=MAX_PATH;
        cbData=MAX_PATH;
    }

    bSuccess=TRUE;

cleanup:

    FreeSid(pSidAdministrators);
    FreeSid(pSidSystem);
    FreeSid(pSidEveryone);
    RegCloseKey(hKey);

    return bSuccess;
}

void MassiveTestLoop(void)
{
    int i;
    char buf[256];

    printf("Press enter to start.\n");
    gets(buf);
    for (i=0; i<10000; i++) {
        if (!ApplyDACLs()) {
            printf("Error, terminating.\n");
            break;
        }
    }
    printf("done.\n");
    gets(buf);
}

void RunServiceInDebugMode(int argc, char *argv[])
{
    if (!ApplyDACLs()) {
        printf("Error setting DACLs\n");
    } else {
        printf("ok.\n");
    }
}

VOID InitServiceThread(VOID)
{
    *StopEvent = CreateEvent (0, TRUE, FALSE, 0);
}

DWORD MainExecutionFunction(void)
{
    HKEY hKey;
    LONG lRv;
    DWORD dwType;
    DWORD dwData;
    DWORD dwSize;
    BOOL bAutoTerminate=FALSE;
    DWORD dwWait;

    lRv=RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                     TEXT(EXT2KEYPARAMS),
                     0,
                     KEY_ALL_ACCESS,
                     &hKey);
    if (lRv==ERROR_SUCCESS) {
        dwSize=sizeof(DWORD);
        RegQueryValueEx(hKey,
                        TEXT(PARAM_AUTOTERMINATE),
                        0,
                        &dwType,
                        (LPBYTE) &dwData,
                        &dwSize);
        if ((lRv==ERROR_SUCCESS)&&(dwType==REG_DWORD)) {
            bAutoTerminate=dwData;
        }
        RegCloseKey(hKey);
    }

    MESSAGEBOX("Initially applying DACLs.");
    ApplyDACLs();
    
    if (bAutoTerminate) {
        MESSAGEBOX("Auto Terminating.");
        KillService();
        return 0;
    }
    *ChangeNotificationEvent = CreateEvent (0, TRUE, FALSE, 0);
    lRv=RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                     TEXT(EXT2KEYUSERS),
                     0,
                     KEY_ALL_ACCESS,
                     &hKey);
    if (lRv!=ERROR_SUCCESS) {
        MessageBox(NULL,"Cannot open registry key " EXT2KEYUSERS ", terminating service.",CP_SERVICE_NAME,MB_SERVICE_NOTIFICATION);
        KillService();
        return 0;
    }

    while (serviceRunning) {
        lRv=RegNotifyChangeKeyValue(hKey,
                                    FALSE,
                                    REG_NOTIFY_CHANGE_NAME|REG_NOTIFY_CHANGE_LAST_SET,
                                    *ChangeNotificationEvent,
                                    TRUE);
        if (lRv!=ERROR_SUCCESS) {
            MessageBox(NULL,"Cannot set notification on registry key " EXT2KEYUSERS ", terminating service.",CP_SERVICE_NAME,MB_SERVICE_NOTIFICATION);
            KillService();
            return 0;
        }

        dwWait=WaitForMultipleObjects(2,
                                      ActionEvents,
                                      FALSE,
                                      INFINITE);
        if (dwWait!=WAIT_OBJECT_0) {
            MESSAGEBOX("StopEvent signaled.");
            break;
        }
        MESSAGEBOX("Registry has changed. Applying DACLs.");
        ApplyDACLs();
        ResetEvent(*ChangeNotificationEvent);
    }
    return 0;
}

// ServiceExecutionThread -
// This is the main thread of execution for the
// service while it is running.
DWORD ServiceExecutionThread(LPDWORD param)
{
    DWORD dwExitCode;

    // call another function to avoid many ExitThread/return pairs
    dwExitCode=MainExecutionFunction();
    ExitThread(dwExitCode);
    return dwExitCode; // to suppress compiler warning
}

VOID StopServiceThread(VOID)
{
    if (ServiceInitialized) {
        SetEvent(*StopEvent);
    }
}
